"""
SAARA CLI Package

Interactive command-line interface for SAARA AI data engine.
This is an optional companion package to saara-ai that provides
interactive wizards, commands, and terminal UI.

Install with: pip install saara-cli
"""

__version__ = "1.6.4"
__author__ = "Kilani Sai Nikhil"
